/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#pragma once

#include "hardware/i2c.h"
#include "pico/stdlib.h"

#define I2C_TIMEOUT_uS        5000      // I2Cタイムアウト
#define POINTER_CONVERT_REG   0x00      // Conversion レジスター
#define POINTER_CONFIG_REG    0x01      // Config レジスター
#define CONVERSION_WAIT_uS    303       // 変換待ち時間(1/3300秒)
#define OPERATION_STATUS_MASK 0x8000    // 変換完了識別用 Configレジスタ 15ビット

// PGA設定毎のフルスケール値
#define SCALE_6144mV 6.144f
#define SCALE_4096mV 4.096f
#define SCALE_2048mV 2.048f
#define SCALE_1024mV 1.024f
#define SCALE_0512mV 0.512f
#define SCALE_0256mV 0.256f

// Config レジスター
#define CONF_OS    (0b1 << 15)      // OS:15    1:変換開始
#define CONF_MUX   (0b000 << 12)    // MUX:14-12  引数から指定する
#define CONF_PGA   (0b000 << 9)     // PGA:11-9   引数から指定する
#define CONF_MODE  (0b1 << 8)       // MODE:8   1:シングルショット
#define CONF_DR    (0b110 << 5)     // DR:7-5   110:3300SPS
#define CONF_CMODE (0b0 << 4)       // CMODE:4  0(デフォルト)
#define CONF_CPOL  (0b0 << 3)       // CPOL:3   0(デフォルト)
#define CONF_CLAT  (0b0 << 2)       // CLAT:2   0(デフォルト)
#define CONF_CQUE  (0b11)           // CQUE:1-0 11:コンパレータOFF/ALERTピン出力なし
#define CONF_BASE  (CONF_OS | CONF_MODE | CONF_DR | CONF_CMODE | \
                   CONF_CPOL | CONF_CLAT | CONF_CQUE)

// MUX選択用 (MUX:14-12)
typedef enum {
    MUX_A0_1 = (0b000 << 12),    // 差動 +:A0 -:A1
    MUX_A0_3 = (0b001 << 12),    // 差動 +:A0 -:A3
    MUX_A1_3 = (0b010 << 12),    // 差動 +:A1 -:A3
    MUX_A2_3 = (0b011 << 12),    // 差動 +:A2 -:A3
    MUX_A0_G = (0b100 << 12),    // シングルエンド:A0
    MUX_A1_G = (0b101 << 12),    // シングルエンド:A1
    MUX_A2_G = (0b110 << 12),    // シングルエンド:A2
    MUX_A3_G = (0b111 << 12)     // シングルエンド:A3
} MUX_e;

// PGA:11-9
typedef enum {
    PGA_2of3 = (0b000 << 9),
    PGA_1 = (0b001 << 9),
    PGA_2 = (0b010 << 9),
    PGA_4 = (0b011 << 9),
    PGA_8 = (0b100 << 9),
    PGA_16 = (0b101 << 9)
} PGA_BIT_e;

// PGAとフルスケール値のマッピング
typedef struct {
    PGA_BIT_e pgabit;        // PGA
    float     scaleValue;    // フルスケール値
} PGA_CONF_t;

// PGA選択用 (PGA_CONF_t配列のインデックス)
typedef enum {
    PGA_6144mv = 0,
    PGA_4096mV,
    PGA_2048mV,
    PGA_1024mV,
    PGA_0512mV,
    PGA_0256mV
} SCALE_e;

void  ADS1015_Init(i2c_inst_t *i, uint8_t a);
float ADS1015_Read(MUX_e mux, SCALE_e pga);
