/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */

#include "ads1015.h"

static i2c_inst_t *i2c;
static uint8_t     addr;

static const PGA_CONF_t pgaConf[] = {
    {PGA_2of3, SCALE_6144mV},
    {PGA_1, SCALE_4096mV},
    {PGA_2, SCALE_2048mV},
    {PGA_4, SCALE_1024mV},
    {PGA_8, SCALE_0512mV},
    {PGA_16, SCALE_0256mV}};

static uint8_t buff[3];

void ADS1015_Init(i2c_inst_t *i, uint8_t a) {
    addr = a;
    i2c = i;
}

/**
 * @brief 指定したレジスタへ書き込み
 */
static void writeRegister(uint8_t reg, uint16_t value) {
    buff[0] = reg;
    buff[1] = value >> 8;
    buff[2] = value & 0xFF;

    i2c_write_timeout_us(i2c, addr, buff, 3, false, I2C_TIMEOUT_uS);
}

/**
 * @brief 指定したレジスタから読み出し
 */
static uint16_t readRegister(uint8_t reg) {
    buff[0] = reg;

    i2c_write_timeout_us(i2c, addr, buff, 1, false, I2C_TIMEOUT_uS);
    i2c_read_timeout_us(i2c, addr, buff, 2, false, I2C_TIMEOUT_uS);

    uint16_t result = ((buff[0] << 8) | buff[1]);
    return result;
}

/**
 * @brief 電圧読み取り
 */
float ADS1015_Read(MUX_e mux, SCALE_e scale) {
    uint16_t config = CONF_BASE | mux | pgaConf[scale].pgabit;

    // 変換開始
    writeRegister(POINTER_CONFIG_REG, config);
    // 変換待ち
    sleep_us(CONVERSION_WAIT_uS);
    while ((readRegister(POINTER_CONFIG_REG) & OPERATION_STATUS_MASK) == 0)
        ;
    // 結果取得
    uint16_t rawdata = readRegister(POINTER_CONVERT_REG);

    // 電圧を計算
    int16_t adcValue = (int16_t)rawdata >> 4;
    if (rawdata >= 0x8000) {    // 負号拡張
        adcValue |= 0xF000;     // 上位4ビットを1
    }
    float result = adcValue * (pgaConf[scale].scaleValue / 2048.0f);    // ±12ビットなので 2^11で割る

    return result;
}