/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>

#include "hardware/gpio.h"
#include "hardware/i2c.h"

#include "ads1015.h"
#include "aqm1602.h"
#include "mcp4726.h"

#include "i2cDev.h"

/**
 * @brief I2Cデバイスを初期化
 */
void initI2C() {
    // i2c0 : ACD,メーター用DAC
    i2c_init(i2c0, I2C0_SPEED_kHz * 1000);
    gpio_set_function(I2C0_SDA_PIN, GPIO_FUNC_I2C);
    gpio_set_function(I2C0_SCL_PIN, GPIO_FUNC_I2C);
    gpio_pull_up(I2C0_SDA_PIN);
    gpio_pull_up(I2C0_SCL_PIN);
    ADS1015_Init(i2c0, ADDR_ADS1015);
    MCP4726_Init(i2c0, ADDR_MCP4726);
    MCP4726_Set(i2c0, ADDR_MCP4726, 0);

    // i2c1 : LCD,TA用DAC
    i2c_init(i2c1, I2C1_SPEED_kHz * 1000);
    gpio_set_function(I2C1_SDA_PIN, GPIO_FUNC_I2C);
    gpio_set_function(I2C1_SCL_PIN, GPIO_FUNC_I2C);
    gpio_pull_up(I2C1_SDA_PIN);
    gpio_pull_up(I2C1_SCL_PIN);
    LCD_Init(i2c1, ADDR_LCD);
    MCP4726_Init(i2c1, ADDR_MCP4726);
    MCP4726_Set(i2c1, ADDR_MCP4726, 0);
}

// ADC /////////////////////////////////////////////////

static float adcRead(MUX_e ch, bool averaging) {
    float   result = 0.0;
    SCALE_e scale = PGA_4096mV;    // 最大3.3v

    result = fabs(ADS1015_Read(ch, scale));

    // 最適なPGA設定で再計測する。
    if (result < SCALE_0256mV) {
        scale = PGA_0256mV;
    } else if (result < SCALE_0512mV) {
        scale = PGA_0512mV;
    } else if (result < SCALE_1024mV) {
        scale = PGA_1024mV;
    } else if (result < SCALE_2048mV) {
        scale = PGA_2048mV;
    }
    result = ADS1015_Read(ch, scale);

    if (averaging) {
        for (int i = 1; i < ABERAGING_NUM; ++i) {
            result += ADS1015_Read(ch, scale);
        }
        result /= ABERAGING_NUM;
    }
    return result;
}

/**  @brief DACの電圧値を計測する */
float readDacV(bool averaging) { return (adcRead(MUX_A0_G, averaging)); }

/** @brief PCの電圧値を計測する */
float readPcV(bool averaging) { return (adcRead(MUX_A1_G, averaging)); }

/** @brief ref電圧を計測する */
float readRefV(bool averaging) { return (adcRead(MUX_A2_G, averaging)); }

/** @brief 出力電圧値を計測する */
float readOutputV(bool averaging) { return (-adcRead(MUX_A2_3, averaging)); }    // a2-a3を計測するので負号を反転する。

// DAC1 /////////////////////////////////////////////////

int16_t setMeterDac(float value) {
    if (value >= INT16_MAX) value = INT16_MAX;
    return MCP4726_Set(i2c0, ADDR_MCP4726, (int16_t)value);
}

// DAC2 /////////////////////////////////////////////////

int16_t setTaDac(int16_t value) {
    int16_t pos = MCP4726_Set(i2c1, ADDR_MCP4726, value);
    sleep_us(DAC_WAIT_US);
    return pos;
}

// LCD /////////////////////////////////////////////////

/**
 * @brief LCDへメッセージを表示する
 */
void lcdPrint(uint8_t row, char *message) {
    LCD_SelectRow(row);
    LCD_Write(message);
}
