/*
 * Copyright (c) 2024 ShinGeTsu Meter.
 *
 * SPDX-License-Identifier: BSD-3-Clause
 */
#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include "hardware/adc.h"
#include "hardware/gpio.h"
#include "hardware/watchdog.h"
#include "pico/bootrom.h"
#include "pico/time.h"

#include "i2cDev.h"
#include "meter.h"

// GPIO
static const uint GPIO_PIN[] = {
    BTN1, BTN2, BTN3, BTN4, BTN5, BTN6, BTN7, BTN8, BTN9,    // ボタン
    TA_RE1, TA_RE2,                                          // ロータリーエンコーダ
};

// 電圧値
float refV = 0.0f;       // 計装アンプのRef電圧
float dacV = 0.0f;       // TA用DAC電圧
float pcV = 0.0f;        // PC電圧
float outputV = 0.0f;    // 計装アンプの出力電圧

// メーター
float   sensGain = 1.0f;      // 感度の補正値
int16_t meterDacValue = 0;    // メーター用DACの設定値
float   readScale = 0.0f;     // リード値：セット位置を基準とした目盛の読み -15 ~ 25 (SAMPLING_uS間の平均値)
int16_t samplingCount = 0;    // SAMPLING_uS間のサンプリング数

// 自動セット用
int16_t autoSetWaitNum = (AUTO_SET_WAIT_uS / SAMPLING_uS);    // 自動セットを待つサンプリング数

// PCのTA
float pcTA = TA_MIN;           // PCのTA値
float taList[TA_HIST_SIZE];    // PCのTAの履歴
float taRange = 0.0f;          // 履歴内の範囲

// トーンアーム
int16_t taDacValue = 0;    // TA用D/Aコンバータの設定値
float   dacTA = TA_MIN;    // DACのTA値

// TAカウンター
float  taCounter = 0.0f;                // TAカウンター
float  counterList[COUNT_HIST_SIZE];    // TAカウンター履歴
size_t counterListIndex = 0;            // TAカウンター履歴のインデックス

// イベントフラグ
bool setEvt = false;          // セット実行
bool taMoveEvt = false;       // トーンアーム操作
bool updateEvt = false;       // LCD表示更新
bool timeExceeded = false;    // タイマー用

// 状態管理
bool   autoSet = false;           // 自動セットモード
bool   counting = false;          // TAカウントモード
bool   pause = false;             // TAカウントを中断
view_t viewMode = VIEW_NORMAL;    // LCD表示モード

/**
 * @brief 電圧からTAを計算する。
 */
static inline float v2TA(float v) {
    return ((v / CONST_V) * (TA_MAX - TA_MIN)) + TA_MIN;
}

/**
 * @brief TAから抵抗値を計算する。
 */
static inline float ta2r(float ta) {
    return CONST_R * (ta - TA_MIN) / (TA_MAX - ta);
}

/**
 * @brief 抵抗値からTAを計算する。
 */
static inline float r2ta(float r) {
    return r / (r + CONST_R) * (TA_MAX - TA_MIN) + TA_MIN;
}

/**
 * @brief 感度を計算する
 */
static float getSens() {
    static float  sensList[SENS_AVE_N];
    static size_t index = 0;
    float         sens = 0.0f;
    if ((11.0f <= readScale) && (readScale <= 27.0f)) {
        float s = GET_SENS(fabs(outputV / (dacV - pcV)));
        sensList[index] = s;
        float sum = 0.0f;
        for (int i = 0; i < SENS_AVE_N; i++) {
            sum += sensList[i];
        }
        sens = sum / SENS_AVE_N;
        index = (index + 1) & (SENS_AVE_N - 1);
    }
    if ((12.5f >= readScale) && (readScale >= 25.0f)) {
        sens = 0.0f;
    }
    return sens;
}

/**
 * @brief TA値による補正係数を計算する。
 */
static float correctioFactor(float ta) {
    float sub400R = ta2r(ta) - 400.0f;    //(100でも1000でも結果は同じ)
    float sub400ta = r2ta(sub400R);
    // 補正なしの場合、TAの変化量とリード幅は比例するので、定抵抗変化時のTA変化量の逆数が補正値となる。
    return 1.0f / (ta - sub400ta);
}

/**
 * @brief TA用DACを変更した際の処理。dacTAと感度補正値を計算する。
 */
void taDacChanged() {
    dacTA = v2TA(dacV);
    float tTA = dacTA;
    if (tTA < CORRECT_TA_MIN) tTA = CORRECT_TA_MIN;
    if (tTA > CORRECT_TA_MAX) tTA = CORRECT_TA_MAX;
    if (tTA > CORRECT_TA_TH) {
        tTA = CORRECT_TA_TH + (tTA - CORRECT_TA_TH) * CORRECT_TA_RATE;
    }
    sensGain = correctioFactor(tTA) / correctioFactor(CORRECT_TA_BASE);
}

/**
 * @brief TAアクションをカウントする。
 */
void countTA(bool memory) {
    static float baseTA = 0;
    if (!pause && counting) {
        if ((baseTA > dacTA)) {
            if (memory) {
                ++counterListIndex;
                counterListIndex = counterListIndex % COUNT_HIST_SIZE;
                counterList[counterListIndex] = taCounter;
            }
            taCounter += (baseTA - dacTA);
            counterList[counterListIndex] = taCounter;
        }
    }
    baseTA = dacTA;
}

/**
 * @brief TAカウントを履歴から復元する
 */
void undoCounter(int num) {
    counterListIndex = (counterListIndex + COUNT_HIST_SIZE + num) % COUNT_HIST_SIZE;
    taCounter = counterList[(counterListIndex + COUNT_HIST_SIZE) % COUNT_HIST_SIZE];
}

/**
 * @brief PCのTA履歴を保存する。履歴からTAの変動範囲を集計する。
 */
void addTaList(float ta) {
    static size_t listIndex = 0;

    taList[listIndex] = ta;
    float maxValue = 0.0f;
    float minValue = 7.0f;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        if (maxValue < taList[i]) {
            maxValue = taList[i];
        }
        if (minValue > taList[i]) {
            minValue = taList[i];
        }
    }
    taRange = fabs(maxValue - minValue);

    listIndex = (listIndex + 1) % TA_HIST_SIZE;
}

/**
 * @brief PCのTA履歴ををクリアする。
 *
 */
void clearTaHist() {
    taRange = 0.0f;
    for (int i = 0; i < TA_HIST_SIZE; i++) {
        taList[i] = pcTA;
    }
}

/**
 * @brief アラームハンドラ
 */
int64_t alarmHandler(alarm_id_t id, void *user_data) {
    timeExceeded = true;
    return 0;
}

/**
 * @brief ボタンイベントを番号毎に管理し実行を判定する。
 */
bool checkBtnAction(size_t eventNo, uint target, uint gpio, uint32_t event) {
    static uint64_t nextAction[] = {0, 0, 0, 0, 0, 0, 0, 0, 0};
    bool            result = false;
    if (target == gpio) {
        uint64_t nowUs = time_us_64();
        result = ((nowUs > nextAction[eventNo]) && (event == GPIO_IRQ_EDGE_FALL));
        nextAction[eventNo] = nowUs + EVENT_WAIT_uS;
    }
    return result;
}

/**
 * @brief GPIO割り込みハンドラ
 */
void irqHandler(uint gpio, uint32_t events) {
    // ロータリーエンコーダー
    if (gpio == TA_RE1 || gpio == TA_RE2) {
        static const int8_t RE_DIR[] = ROTARY_DIRECTION;
        static uint8_t      taRE1;
        static uint8_t      taRE2;
        static size_t       reIndex = 0;

        if (setEvt) {
            setEvt = false;
        } else {
            uint8_t state = (events == GPIO_IRQ_EDGE_RISE) ? 1 : 0;
            taRE1 = (gpio == TA_RE1) ? state : taRE1;
            taRE2 = (gpio == TA_RE2) ? state : taRE2;

            size_t lastIndex = reIndex;
            reIndex = ((reIndex << 2) + (taRE1 << 1) + taRE2) & 15;
            if (lastIndex != reIndex) {
                int8_t delta = RE_DIR[reIndex];
                if (delta != 0) {
                    taDacValue += delta;
                    taMoveEvt = true;
                }
            }
        }
    }

    // セットボタン Event0
    else if (checkBtnAction(0, BUTTON_SET, gpio, events)) {
        setEvt = true;
        pause = false;
    }

    // カウントモードボタン Event1
    else if (checkBtnAction(1, BUTTON_COUNT, gpio, events)) {
        if (!counting) {
            pause = false;
            countTA(false);    // 現在のTAをBaseTAとするため
            counting = true;
        } else {
            counting = false;
        }
        updateEvt = true;
    }

    // 自動モードボタン Event2
    else if (checkBtnAction(2, BUTTON_AUTO, gpio, events)) {
        autoSet = !autoSet;
        updateEvt = true;
    }

    // 電源表示ボタン Event3
    else if (gpio == BUTTON_POWER) {
        viewMode = VIEW_NORMAL;
        updateEvt = true;
    }

    // 表示モードボタン Event4
    else if (checkBtnAction(4, BUTTON_VIEW, gpio, events)) {
        viewMode = (++viewMode) % 4;
        setEvt = false;
        updateEvt = true;
    }

    // クリアボタン Event5
    else if (checkBtnAction(5, BUTTON_CLEAR, gpio, events)) {
        if (viewMode == VIEW_CONF) {
            autoSetWaitNum = 0;
        } else {
            undoCounter(1);
            countTA(false);    // 現在のTAをBaseTAとするため
            taCounter = 0.0f;
            counterList[counterListIndex] = taCounter;
            pause = false;
        }
        updateEvt = true;
    }

    // 戻る Event6
    else if (checkBtnAction(6, BUTTON_BACK, gpio, events)) {
        if (viewMode == VIEW_CONF) {
            autoSetWaitNum -= AUTO_SET_ADJ_STEP;
            if (autoSetWaitNum < 0) {
                autoSetWaitNum = 0;
            }
        } else {
            undoCounter(-1);
        }
        updateEvt = true;
    }

    // 進む Event7
    else if (checkBtnAction(7, BUTTON_FORWARD, gpio, events)) {
        if (viewMode == VIEW_CONF) {
            if ((autoSetWaitNum * SAMPLING_uS) < (5 * 1000 * 1000)) {    // 最大5秒
                autoSetWaitNum += AUTO_SET_ADJ_STEP;
            }
        } else {
            undoCounter(1);
        }
        updateEvt = true;
    }

    // セットボタンその２ Event8
    else if (checkBtnAction(8, BUTTON_SET2, gpio, events)) {
        setEvt = true;
        pause = false;
    }
}

/**
 * @brief GPIO初期化
 */
void initIO() {
    // LED
    gpio_init(LED_PIN);
    gpio_set_dir(LED_PIN, GPIO_OUT);

    // USB接続チェック用
    gpio_init(VBUS_SENSE_PIN);
    gpio_set_dir(VBUS_SENSE_PIN, GPIO_IN);

    // 電源チェック用
    adc_init();
    adc_gpio_init(VSYS_ADC_PIN);

    // GPIO
    uint size = sizeof(GPIO_PIN) / sizeof(uint);
    for (int i = 0; i < size; ++i) {
        gpio_init(GPIO_PIN[i]);
        gpio_set_dir(GPIO_PIN[i], GPIO_IN);
        gpio_pull_up(GPIO_PIN[i]);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_FALL, true, irqHandler);
        gpio_set_irq_enabled_with_callback(GPIO_PIN[i], GPIO_IRQ_EDGE_RISE, true, irqHandler);
    }
}

/**
 * @brief LCDを更新する。
 */
void update() {
    int view = viewMode;
    if (!gpio_get(BUTTON_POWER)) {
        view = VIEW_CONF;
    }
    char       *label;
    static char message[30];

    switch (view) {
        case VIEW_NORMAL:    // 通常表示
            if (setEvt) {
                label = "SET";
            } else {
                char labelStatus[] = {0x20, 0x20, 0x20, 0};
                label = labelStatus;
                if (autoSet) {
                    labelStatus[2] = 'a';
                }
                if (counting) {
                    labelStatus[1] = 'c';
                    if (pause) {
                        labelStatus[0] = '!';
                    }
                }
            }
            sprintf(message, "%4.2f %5.2f   %s", pcTA, pcTA - dacTA, label);
            lcdPrint(1, message);
            sprintf(message, "%4.2f %5.1f %5.2f", dacTA, readScale, taCounter);
            lcdPrint(2, message);
            break;
        case VIEW_TA:    // 較正モード TA詳細
            sprintf(message, "%6.4f %+6.4f   ", pcTA, pcTA - dacTA);
            lcdPrint(1, message);
            sprintf(message, "%6.4f %4.2f%5.1f", dacTA, sensGain, (pcTA < PC_TA_MAX) ? getSens() : 0.0);
            lcdPrint(2, message);
            break;
        case VIEW_ADC:    // 較正モード 電圧詳細
            sprintf(message, "%7.5f %7.5f ", pcV, refV);
            lcdPrint(1, message);
            sprintf(message, "%7.5f %7.5f ", dacV, outputV);
            lcdPrint(2, message);
            break;
        case VIEW_CONF:    // 電源状態表示
            // 電源電圧計測
            adc_select_input(VSYS_ADC_NO);
            float vsys = adc_read() * (3.3 * 3.0 / 4096.0);
            if (gpio_get(VBUS_SENSE_PIN)) {    // USBによる給電
                label = "USB :";
            } else {
                label = "BATT:";
                vsys += 0.2;    // ショットキーバリアダイオード分
            }
            sprintf(message, "%s%4.2fv      ", label, vsys);
            lcdPrint(1, message);
            sprintf(message, "auto set:%3.1fsec ", (SAMPLING_uS / 1000.0 / 1000.0) * (float)autoSetWaitNum);
            lcdPrint(2, message);
            break;
    }
}

/**
 * @brief 状態を確認する。
 */
void checkState() {
    static uint16_t outRangeNum = 0;

    if (autoSet && (viewMode == VIEW_NORMAL)) {
        // TAの変動幅がTA_BLOCK_RANGEを超過している場合カウントを停止
        if ((taRange > TA_PAUSE_RANGE)) {
            pause = true;
        }

        // 自動セット
        float meterScale = readScale + METER_SCALE_SET;
        if ((pcTA < PC_TA_MAX) &&                                           //
            ((meterScale <= 0) || (meterScale >= (METER_SCALE_MAX + 2)))    //
        ) {
            outRangeNum++;
            if (autoSetWaitNum <= outRangeNum) {
                setEvt = true;
                outRangeNum = 0;
            }
        } else {
            outRangeNum = 0;
        }
    }
}

/**
 * @brief 針をセット位置へ移動する。
 */
void setNeedle() {
    // セット位置計算の補正用
    // dacValue = (pcV / constV) * fa + fb;
    // (vRate:pcV/ConstV,value:dacValue)
    static float   p1vRate = 0.0f;
    static int16_t p1value = 0;
    static float   p2vRate = 1.0f;
    static int16_t p2value = DAC_MAX_VALUE - 50.0f;
    static float   fa = DAC_MAX_VALUE - 50.0f;
    static float   fb = 0.0f;

    // SET位置へジャンプ
    pcV = readPcV(true);
    pcTA = v2TA(pcV);
    taDacValue = setTaDac((pcV / CONST_V) * fa + fb);

    // 微調整
    dacV = readDacV(true);
    bool dir = pcV > dacV;
    while (setEvt) {
        int16_t tValue = taDacValue;
        float   tDacV = dacV;

        taDacValue = setTaDac(taDacValue + (dir ? +1 : -1));
        dacV = readDacV(true);
        if (tValue == taDacValue) {    // 設定可能範囲を超過
            setEvt = false;
            break;
        }

        if (dir != (pcV > dacV)) {                         // セット位置を通り過ぎたら終了
            if (fabs(pcV - tDacV) < fabs(pcV - dacV)) {    // 通り過ぎる前の方が近かったら戻す
                taDacValue = setTaDac(tValue);
                dacV = readDacV(true);
            }
            setEvt = false;
        }
    }
    taDacChanged();
    readScale = 0.0f;

    // SET位置計算式を最適化
    bool optimize = false;
    // 近いほうを入れ替え
    if (abs(p1value - taDacValue) < abs(p2value - taDacValue)) {
        // 入れ替え後の差がDAC設定範囲の約5%以下ならば入れ替えない
        if (abs(p2value - taDacValue) > 200) {
            p1vRate = pcV / CONST_V;
            p1value = taDacValue;
            optimize = true;
        }
    } else {
        // 入れ替え後の差がDAC設定範囲の約5%以下ならば入れ替えない
        if (abs(p1value - taDacValue) > 200) {
            p2vRate = pcV / CONST_V;
            p2value = taDacValue;
            optimize = true;
        }
    }
    if (optimize == true) {
        fa = (p1value - p2value) / (p1vRate - p2vRate);
        fb = p1value - p1vRate * (p1value - p2value) / (p1vRate - p2vRate);
    }
}

/**
 * @brief メーターを駆動する。
 */
void driveNeedle() {
    samplingCount = 0;
    float meterVSum = 0.0f;
    float outputVSum = 0.0f;
    timeExceeded = false;
    add_alarm_in_us(SAMPLING_uS, alarmHandler, NULL, false);
    while (!timeExceeded) {
        ++samplingCount;
        if (pcTA > PC_TA_MAX) {
            sleep_ms(10);    // 缶を持っていない場合何もしない(節電)
        } else {
            float v = readOutputV(false);    // セット位置との差分の電圧
            outputVSum += v;
            float meterV = v * sensGain + 1.0f;    // セット位置を1として計算する(実電圧の考慮不要)
            meterVSum += meterV;
            if (meterV < 0.0f) meterV = 0.0f;                                                    // 負電圧を出力できない
            float pos = meterV * (DAC_MAX_VALUE) / (METER_SCALE_MAX + 2) * (METER_SCALE_SET);    // +2:目盛範囲を超えて振れるようにする
            meterDacValue = setMeterDac(pos);
        }
    }

    // 目盛位置を計算
    outputV = outputVSum / samplingCount;
    readScale = meterVSum / samplingCount * METER_SCALE_SET - METER_SCALE_SET;
    if (readScale > 0) readScale = readScale * METER_ADJ;    // 目盛値の補正
    if (readScale > (99.0f)) readScale = 99.0f;              // LCDの表示桁数に収める
    if (readScale < (-99.0f)) readScale = -99.0f;            // LCDの表示桁数に収める
}

/**
 * @brief メイン
 */
int main() {
    sleep_ms(100);
    initIO();
    initI2C();
    sleep_ms(100);

    autoSet = false;
    setEvt = false;
    taMoveEvt = false;
    updateEvt = true;

    clearTaHist();
    for (size_t i = 0; i < COUNT_HIST_SIZE; i++) {
        counterList[i] = 0.0f;
    }

    // 設定モードを表示
    viewMode = VIEW_CONF;
    update();
    uint64_t waitUs = time_us_64() + POWER_VIEW_WAIT_uS;
    while (viewMode != VIEW_NORMAL) {
        if (time_us_64() > waitUs) {
            viewMode = VIEW_NORMAL;
            break;
        }
        // 自動セット待ち時間を調整した場合、表示時間を延ばす
        else if (!gpio_get(BUTTON_BACK) ||
                 !gpio_get(BUTTON_FORWARD) ||
                 !gpio_get(BUTTON_CLEAR))    //
        {
            waitUs = time_us_64() + POWER_VIEW_WAIT_uS;
            update();
        } else if (!gpio_get(BUTTON_SET) && gpio_get(VBUS_SENSE_PIN)) {
            lcdPrint(1, "Device Firmware ");
            lcdPrint(2, "Update          ");
            reset_usb_boot(0, 0);
        }
        sleep_ms(1);
    }

    uint64_t netxUpdateUs = 0;
    refV = readRefV(true);
    readScale = 0.0f;

    while (true) {
        checkState();

        // トーンアームイベント
        if (taMoveEvt) {
            taDacValue = setTaDac(taDacValue);
            dacV = readDacV(true);
            taDacChanged();
            countTA(false);    // トーンアーム操作によるカウントは履歴へ含めない
            update();
            taMoveEvt = false;
        }
        // セットイベント
        else if (setEvt) {
            gpio_put(LED_PIN, true);
            update();
            setNeedle();
            readScale = 0.0f;
            clearTaHist();
            countTA(true);
            netxUpdateUs = time_us_64() + UPDATE_uS;    // 次の表示更新まで"SET"を表示する
            gpio_put(LED_PIN, false);
        }
        // 表示更新イベント
        else if (updateEvt) {
            update();
            updateEvt = false;
        }

        // イベント外
        else {
            driveNeedle();

            // PCの電圧をサンプリング
            pcV = readPcV(viewMode == VIEW_ADC);
            pcTA = v2TA(pcV);
            addTaList(pcTA);

            // 較正モードの場合 TA用DACとRefVをサンプリング
            if (viewMode == VIEW_ADC) {
                refV = readRefV(true);
                dacV = readDacV(true);
                taDacChanged();
            }

            // UPDATE_uS以上経過していたらLCDを更新
            uint64_t nowUs = time_us_64();
            if (nowUs >= netxUpdateUs) {
                update();
                netxUpdateUs = nowUs + UPDATE_uS;
            }
        }
    }
    return 0;
}